
# THIS FILE IS GENERATED FROM SETUP.PY
short_version = '0.9.0'
version = '0.9.0'
full_version = '0.9.0'
git_revision = 'a8889192d705e86397acf1473fa3f4336b0b6816'
release = True

if not release:
    version = full_version